(function () {
	'use strict';
	'format amd';

	function __util(angular, inview) {
		angular.module('cevo.util', ['angular-inview', 'ui.router']).provider('cevo', function () {
			var util = {
				objectify: function objectify(arr, obj) {
					for (var i in arr) {
						arr[i] = new obj(arr[i]);
					}return arr;
				},

				object: {
					size: function size(obj) {
						var size = 0,
						    key;

						for (key in obj) {
							if (obj.hasOwnProperty(key)) size++;
						}

						return size;
					}
				},

				getAttributes: function getAttributes(obj) {
					var attributes = {};

					for (var i in obj) {
						if (typeof obj[i] !== 'function') attributes[i] = obj[i];
					}

					return attributes;
				}
			};

			var config = {
				notification: {
					defaultTimeout: 7000,
					defaultType: 'global'
				},
				oauth: {
					clientId: '',
					activationUrl: ''
				},
				env: 'production',
				endpoint: '',
				util: util
			};

			return {
				setEnvironment: function setEnvironment(val) {
					config.env = val;
				},

				setEndpoint: function setEndpoint(val) {
					config.endpoint = val;
				},

				setOAuth: function setOAuth(val) {
					config.oauth = val;
				},

				notification: {
					setDefaultTimeout: function setDefaultTimeout(val) {
						config.notification.defaultTimeout = val;
					},

					setDefaultType: function setDefaultType(val) {
						config.notification.defaultType = val;
					}
				},

				$get: function $get() {
					return config;
				}
			};
		});

		return 'cevo.util';
	}

	module.exports = __util(require('angular'), require('angular-inview'), require('@uirouter/angularjs'));
})();
angular.module('cevo.util').filter('encodeNewlines', ["$sce", function ($sce) {
    return function (msg) {
        msg = (msg + '').replace(/\\r\\n/g, "\r\n");
        msg = (msg + '').replace(/\\t/g, "\t");

        return $sce.trustAsHtml(msg);
    };
}]);
angular.module('cevo.util').filter('nl2br', ["$sce", function ($sce) {
    return function (msg, is_xhtml) {
        var is_xhtml = is_xhtml || true;
        var breakTag = '<br />';
        var msg = (msg + '').replace(/([^>\r\n]?)(\r\n|\n\r|\r|\n)/g, '$1' + breakTag);
        return $sce.trustAsHtml(msg);
    };
}]);
angular.module('cevo.util').filter('ordinal', ["$filter", function ($filter) {
	var suffixes = ["th", "st", "nd", "rd"];

	return function (number) {
		var suffix = $filter('ordinalSuffix')(number);
		return number + suffix;
	};
}]);
angular.module('cevo.util').filter('ordinalSuffix', ["$filter", function ($filter) {
	var suffixes = ["th", "st", "nd", "rd"];

	return function (number) {
		var relevantDigits = number < 30 ? number % 20 : number % 30;
		var suffix = relevantDigits <= 3 ? suffixes[relevantDigits] : suffixes[0];

		return suffix;
	};
}]);
angular.module('cevo.util').filter('SplitArray', ["Logger", function (Logger) {
	return function (items, columns) {
		if (!items) return [items];

		if (items.length <= columns) {
			return [items];
		};

		var rowsNum = Math.ceil(items.length / columns);

		var rowsArray = new Array(rowsNum);

		for (var i = 0; i < rowsNum; i++) {
			var columnsArray = new Array(columns);
			for (var j = 0; j < columns; j++) {
				var index = i * columns + j;

				if (index < items.length) {
					columnsArray[j] = items[index];
				} else {
					break;
				}
			}
			rowsArray[i] = columnsArray;
		}

		Logger.debug('Finished splitting', rowsArray);

		return rowsArray;
	};
}]);
angular.module('cevo.util').filter('stripTags', function () {
	return function (string) {
		return string ? String(string).replace(/<[^>]+>/gm, '') : '';
	};
});
angular.module('cevo.util').filter('timezoneAbbreviation', ["Logger", function (Logger) {
	return function (string, concat) {
		if (typeof concat === 'undefined') concat = true;

		//Logger.debug('timezone', string, concat);

		var date = Date.now();
		if (typeof Intl === 'undefined') return (concat ? string : '') + ' LT';

		var idate = new Intl.DateTimeFormat('en-US', { 'timeZoneName': 'short' }).format(date);
		return (concat ? string : '') + ' ' + idate.split(' ').reverse()[0];
	};
}]);
angular.module('cevo.util').filter('ToArray', ["Logger", "cevo", function (Logger, cevo) {
	return function (items) {
		var array = [];

		//go through each filter
		//Logger.debug('To Array', cevo);
		if (cevo.util.object.size(items) <= 0) return array;

		for (var item in items) {
			item = items[item];
			array.push(item);
		}

		return array;
	};
}]);
var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

angular.module('cevo.util').service('Form', ["Logger", function (Logger) {
	/*
 - In order to cleanly send files via ajax requests, we need to use the FormData object.
 - If we use FormData objects, values within FormData that are objects are not encoded, and thus must be stringified with angular.toJson.
 - If the form's data contain no files, we don't need to use multipart/form-data as the request type, and can instead use application/json 
 and simply encode the data passed into this method.
 - Likewise, a FormData object can be passed directly into this method and it will be immediately returned.
 - Adding some code that allows us to pass objects with methods into formdata by stripping out anything that is a function
 */
	this.errors = {};

	this.clean = function (form) {
		var formData;
		var sendingFile = false;

		if (form instanceof FormData) formData = form;else {
			for (var i in form) {
				if (form[i] instanceof File) {
					sendingFile = true;
					break;
				} else if (angular.isFunction(form[i])) delete form[i];
			}

			if (sendingFile) {
				formData = new FormData();
				for (var i in form) {
					var data = form[i];
					if (angular.isFunction(data)) continue;

					if (!(data instanceof File) && (typeof data === 'undefined' ? 'undefined' : _typeof(data)) === 'object') data = angular.toJson(data);

					formData.append(i, data);
				}
			} else formData = form;
		}

		return {
			'data': formData,
			'contentType': sendingFile ? undefined : 'application/json' //undefined will let the browser decide, but it's shitty at determining what is JSON
		};
	};

	this.setErrors = function (errors) {
		Logger.debug('Setting form errors', errors);
		this.errors = errors;
	};
}]);
//https://medium.com/@skosno/server-error-handing-in-angularjs-forms-c5fd38ccf0fc
angular.module('cevo.util').service('FormValidation', ["Logger", "Notification", function (Logger, Notification) {
	var _this = this;

	var formValidation = this;

	this.applyErrors = function (form, data) {
		Logger.debug('form validation', 'Applying errors... ', form, data);

		if (!data || !data.data || !data.data.errors) return Notification.error('There was an unknown error with the request.');

		_this.reset(form);

		var errors = data.data.errors;

		Logger.debug('form validation', 'foreach ', errors);
		angular.forEach(errors, _this.applyError.bind(null, form));

		for (var i in errors) {
			//var error = errors[i];
			//Logger.debug('form', 'Looping errors, saw', i, type);
			//angular.forEach(type, this.applyError.bind(null, form))
		}
	};

	this.applyError = function (form, error) {
		var code = error.code,
		    field = error.field || error.tag,
		    message = error.message,
		    api;

		Logger.debug('form validation', 'Applying error', form, error, code, field, message);

		if (!form.$api) _this.reset(form);

		api = form.$api;
		Logger.debug('form validation', 'got api', api);

		//field is not able to handle API errors
		if (!form[field] || !form[field].$registerError) {
			Logger.debug('form validation', 'did not see form field, cant register error', code, field, message, form[field]);
			if (!api.$errors) api.$errors = [];

			api.$errors.push(error);
			_this.setInvalid(form);
		} else {
			Logger.debug('form validation', 'Applied new error', form[field], form[field].$registerError, field, message);
			form[field].$registerError(error);
		}

		return true;
	};

	this.reset = function (form) {
		form.$api = {
			$errors: [],
			$invalid: false,
			$valid: true
		};

		form.$setPristine();
		form.$setUntouched();

		for (var i in form) {
			if (form[i] && form[i].$errors) {
				Logger.debug('form validations', 'resetting form object errors', form[i]);
				form[i].$errors = [];
				form[i].$valid = true;
				form[i].$invalid = false;
			}
		}
	};

	this.setInvalid = function (form) {
		if (!form.$api) this.reset(form);

		form.$api.$invalid = true;
		form.$api.$valid = false;
	};

	this.setValid = function (form) {
		if (!form.$api) this.reset(form);

		form.$api.$invalid = false;
		form.$api.$valid = true;
	};
}]);
angular.module('cevo.util').service('Framework', function () {
	this.header = true;
	this.subnav = true;
	this.content = true;
	this.footer = true;
	this.bare = false;
	this.backgroundImage = '';
	this.backgroundColor = '';
	this.title = '';
});
var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

angular.module('cevo.util').service('Listener', ["Logger", function (Logger) {
	var obj = this;

	this.id = 0;

	var listeners = {};

	this.$registerListener = function (action) {
		listeners[action] = [];

		//Logger.debug('listener', 'registered', action, listeners);
	};

	this.$on = function (actions, func, tag) {
		//Logger.debug('Registering new listener', actions, func);
		if ((typeof actions === 'undefined' ? 'undefined' : _typeof(actions)) !== 'object' && typeof actions != 'array') actions = [actions];

		if (!tag) tag = Math.random();

		for (var i in actions) {
			var action = actions[i];
			if (!listeners[action]) return false;

			listeners[action].push({ tag: tag, func: func });

			//Logger.debug('notifyer, Registered new listener', actions, listeners);
		}

		return tag;
	};

	//return true if a listener was successfully unregistered, false if nothing was found
	this.$off = function (actions, tag) {
		var success = false;
		if (!angular.isObject(actions) && !angular.isArray(actions)) //typeof actions !== 'object' && typeof actions != 'array')
			actions = [actions];

		for (var a in actions) {
			var action = actions[a];
			for (var i in listeners[action]) {
				var listener = listeners[action][i];

				if (listener.tag && listener.tag == tag) {
					//Logger.debug('Deregistering', action, tag);
					delete listeners[action][i];
					success = true;
					break;
				}
			}
		}

		return success;
	};

	this.$notify = function (action) {
		//Logger.debug('notifying ', listeners, action);

		if (!listeners[action]) return false;

		for (var i in listeners[action]) {
			//Logger.debug('Calling', listeners[action][i].tag, listeners[action][i].func);
			listeners[action][i].func.call();
		}
	};
}]);
angular.module('cevo.util').service('Logger', ["$log", "cevo", function ($log, cevo) {
	this.logs = [];
	this.log = function () {
		for (var i in arguments) {
			this.logs.push({ 'type': 'log', 'message': arguments[i], 'time': new Date().toUTCString() });
		}$log.log.apply(this, arguments);
	};

	this.debug = function () {
		for (var i in arguments) {
			this.logs.push({ 'type': 'debug', 'message': arguments[i], 'time': new Date().toUTCString() });
		}if (cevo.env != 'production') $log.debug.apply(this, arguments);
	};

	this.error = function () {
		for (var i in arguments) {
			this.logs.push({ 'type': 'error', 'message': arguments[i], 'time': new Date().toUTCString() });
		}$log.error.apply(this, arguments);
	};
}]);
angular.module('cevo.util').service('MetaTags', ["$http", "$window", "cevo", "Logger", "$sce", function ($http, $window, cevo, Logger, $sce) {
    var obj = this;

    obj.defaults = {
        title: '',
        description: '',
        og: {},
        twitter: {},
        height: '300',
        width: '400'
    };

    obj.reset = function () {
        obj.title = obj.defaults.title;
        obj.description = obj.defaults.description;
        obj.og = obj.defaults.og;
        obj.twitter = obj.defaults.twitter;
        obj.height = obj.defaults.height;
        obj.width = obj.defaults.width;
    };

    obj.getOrigin = function () {
        return $window.location.origin;
    };

    obj.reset();

    return obj;
}]);
angular.module('cevo.util').service('MIDIHelper', ["Logger", "$q", function (Logger, $q) {
	this.logs = [];
	this.MIDI = window.MIDI;
	this.MIDIAccess;
	this.outputs = [];

	var MIDIHelper = this;

	if (!this.MIDI) Logger.error('WARNING: MIDI Unavailable!');

	this.requestAccess = function (message) {
		Logger.debug('Requesting access to MIDI devices');
		if (navigator.requestMIDIAccess) {
			return navigator.requestMIDIAccess({
				sysex: false
			}).then(function (MIDIAccess) {
				Logger.debug('triggering onrequest success');
				return MIDIHelper.onRequestSuccess(MIDIAccess);
			}, function () {
				return MIDIHelper.onRequestFailure();
			});
		} else {
			Logger.error('ERROR: MIDI Access Request Failed!');
			return $q.reject();
		}
	};

	this.onRequestSuccess = function (MIDIAccess) {
		Logger.debug('MIDI Access Request Successful', MIDIAccess);

		MIDIHelper.MIDIAccess = MIDIAccess;
		MIDIHelper.storeOutputDevices();

		Logger.debug('resolving promise');
		return $q.resolve();
	};

	this.onRequestFailure = function () {
		Logger.error('ERROR: MIDI Access Request Failed');

		return $q.reject();
	};

	this.storeOutputDevices = function () {
		this.outputs = this.MIDIAccess.outputs;
	};

	this.sendNote = function (note, length) {
		this.outputs.forEach(function (output, port) {
			Logger.debug('output: ' + port, output);
			output.send([0x90, note, 0x7f]); //noteOn, velocity(?), note byte (G9)
			output.send([0x80, note, 0x7f]); //noteOff, velocity(?), note byte (G9)
		});
	};
}]);
angular.module('cevo.util').service('Nav', ["$state", function ($state) {
	this.nav = {};
	this.current = {
		state: $state.current,
		controller: ''
	};

	this.setCurrentState = function (state) {
		this.current.state = state;
		this.setCurrentController();
	};

	this.setCurrentController = function () {
		var parts = this.current.state.name.split('.');
		var site = parts[0];
		this.current.controller = parts[1] || site;
		this.current.page = parts[2] || '';
	};

	this.getCurrentNav = function () {
		return this.nav[this.current.controller] || [];
	};

	this.seed = function () {
		var states = $state.get();

		for (var i in states) {
			var state = states[i];

			if (state.visibility && state.visibility == 'hidden') continue;

			var parts = state.name.split('.');
			var site = parts[0];

			if (!site) continue;

			var controller = parts[1] || site;
			var page = parts[2] || '';

			var testing;

			if (!this.nav[controller]) {
				this.nav[controller] = {
					'nav': [], //use an array here so we can iterate cleanly
					'name': state.name,
					'controller': state.controller,
					'url': state.url,
					'title': controller,
					'class': ''
				};
			}

			if (page) {
				this.append(controller, {
					'nav': [], //allow us to add a subnav, use an array again to easily iterate over it in the view
					'url': this.nav[controller].url + state.url,
					'title': page,
					'name': state.name,
					'class': 'capitalize'
				});
			}
		}
	};

	this.prepend = function (controller, nav, position) {
		if (!controller || !this.nav[controller] || !nav.title && !nav.class || this.exists(controller, nav.tag || nav.title || '')) return false;

		if (position) return this.add(controller, nav, position);else return this.nav[controller].nav.unshift(nav);
	};

	this.append = function (controller, nav, position) {
		if (!controller || !this.nav[controller] || !nav.title && !nav.class || this.exists(controller, nav.tag || nav.title || '')) return false;

		if (position) return this.add(controller, nav, position);else return this.nav[controller].nav.push(nav);
	};

	this.add = function (controller, nav, position, replace) {
		if (!controller || !this.nav[controller] || !nav.title && !nav.class) return false;

		if (typeof replace === 'undefined') //replace by default
			replace = true;

		if (replace) {
			var index = this.exists(controller, nav.tag || nav.title || '');
			if (index) return this.nav[controller].nav[index] = nav;
		}

		return this.nav[controller].nav.splice(position, 0, nav);
	};

	//returns the index of the existing nav item
	this.exists = function (controller, tag) {
		if (!controller || !this.nav[controller]) return false;

		for (var i in this.nav[controller].nav) {
			var item = this.nav[controller].nav[i];
			if ((item.tag || item.title || '') == tag) return i;
		}

		return false;
	};

	this.unset = function (controller, tag) {
		if (!controller || !this.nav[controller]) return false;

		for (var i in this.nav[controller].nav) {
			var item = this.nav[controller].nav[i];
			if ((item.tag || item.title || '') == tag) return this.nav[controller].nav.splice(i, 1);
		}

		return false;
	};
}]);
angular.module('cevo.util').service('Notification', ["$timeout", "cevo", "Logger", function ($timeout, cevo, Logger) {
	this.notifications = [];
	this.listeners = [];

	this.alert = function (message, style, type, duration) {
		if (angular.isObject(message)) {
			for (var i in message) {
				var notification = message[i];
				this.alert(notification.message, style, type, duration);
			}

			return;
		}

		if (!type) type = cevo.notification.defaultType;

		if (!duration) duration = cevo.notification.defaultTimeout;

		this.notifications.push({ 'type': type, 'style': style, 'message': message, 'time': new Date().toUTCString() });
		this.notify();

		var e = this.notifications;
		$timeout(function () {
			e.shift();
		}, duration);
	};

	this.error = function (message, type, duration) {
		Logger.debug('Alerting new error message');
		this.alert(message, 'error', type, duration);
	};

	this.success = function (message, type, duration) {
		Logger.debug('Alerting new success message');
		this.alert(message, 'success', type, duration);
	};

	this.listen = function (callback) {
		Logger.debug('Attaching new listener');
		this.listeners.push(callback);
	};

	this.notify = function () {
		Logger.debug('Notifying listeners', this.listeners);

		for (var i in this.listeners) {
			this.listeners[i].call();
		}
	};
}]);
angular.module('cevo.util').factory('Rotator', ["$timeout", "Logger", function ($timeout, Logger) {
	function Rotator() {
		var delay = 5000;
		var rotateTimeout;

		var obj = {
			currentItem: {},
			currentIndex: -1,
			length: 0,
			items: [],

			config: function config(i, d, run) {
				if (!i || i.length <= 0) return false;

				run = run !== false;
				obj.items = i;

				obj.length = obj.items.length;

				if (d) delay = d;

				//Logger.debug('rotator', 'finished config', i, d, run);

				if (run) obj.run();
			},

			run: function run() {
				obj.next(-1);
			},

			stop: function stop() {
				$timeout.cancel(rotateTimeout);
			},

			next: function next(index) {
				if (typeof index === 'undefined') index = obj.currentIndex;

				//Logger.debug('rotator next', index, obj.items.length, obj.items)

				if (++index >= obj.items.length) index = 0;

				obj.selectItem(index);
			},

			previous: function previous(index) {
				if (typeof index === 'undefined') index = obj.currentIndex;

				//Logger.debug('rotator previous', index, obj.items.length, obj.items)
				if (--index < 0) index = obj.items.length - 1;

				obj.selectItem(index);
			},

			selectItem: function selectItem(index) {
				if (rotateTimeout) $timeout.cancel(rotateTimeout);

				//swLogger.debug('rotator', 'selecting item', index, obj.items, obj.items[index]);

				obj.currentItem = obj.items[index];
				obj.currentIndex = index;

				rotateTimeout = $timeout(function () {
					obj.next(index);
				}, delay);
			}
		};

		return obj;
	}

	return Rotator;
}]);
angular.module('cevo.util').factory('Socket', ["Logger", "$rootScope", function (Logger, $rootScope) {
	function Socket() {
		var socket;
		var name;

		return {
			connect: function connect(host, transports, opts) {
				//Logger.log('connecting to '+host);

				if (typeof io === 'undefined') {
					return;
				}

				name = host;

				var opts = opts || {};
				if (!transports) opts.transports = ['websocket'];

				socket = io.connect(host, opts);

				//Logger.log(name+' connected: '+socket.json.connected);
				socket.on('disconnect', function () {
					if (socket.json) Logger.debug(name + ' connected: ' + socket.json.connected);else Logger.debug(name + ' connected: ' + socket.socket.connected);
				});
				socket.on('connect', function () {
					if (socket.json) Logger.debug(name + ' connected: ' + socket.json.connected);else Logger.debug(name + ' connected: ' + socket.socket.connected);
				});
			},

			on: function on(eventName, callback) {
				Logger.debug(name + ': Setting listener for ' + eventName);

				if (socket) {
					socket.on(eventName, function () {
						var args = arguments;
						$rootScope.$apply(function () {
							callback.apply(socket, args);
						});
					});
				}
			},

			once: function once(eventName, callback) {
				Logger.debug(name + ': Setting listener once for ' + eventName);

				if (socket) {
					socket.once(eventName, function () {
						var args = arguments;
						$rootScope.$apply(function () {
							callback.apply(socket, args);
						});
					});
				}
			},

			off: function off(eventName) {
				Logger.debug(name, ': Removing listener for ', eventName, socket);
				if (socket && socket.removeAllListeners) socket.removeAllListeners(eventName);
			},

			emit: function emit(eventName, data, callback) {
				//console.log('sending '+data+' to '+name) 
				if (socket && (socket.json && socket.json.connected || socket.socket && socket.socket.connected)) {
					//$log.log('emitting ...'+eventName);
					socket.emit(eventName, data, function () {
						//$log.log('done emitting...');
						var args = arguments;
						$rootScope.$apply(function () {
							if (callback) {
								callback.apply(socket, args);
							}
						});
					});
				}
			},

			disconnect: function disconnect() {
				if (socket.json) Logger.debug(name + ' connected: ' + socket.json.connected);else Logger.debug(name + ' connected: ' + socket.socket.connected);

				if (socket.json.connected) socket.disconnect();

				if (socket.json) Logger.debug(name + ' connected: ' + socket.json.connected);else Logger.debug(name + ' connected: ' + socket.socket.connected);
			}
		};
	}

	return Socket;
}]);
angular.module('cevo.util').directive('cActivateOnScroll', ["$q", "$window", "$timeout", "Logger", function ($q, $window, $timeout, Logger) {
	return {
		restrict: 'A',
		scope: {
			'activeClass': '@cActivateOnScroll',
			'id': '@cActivateId'
		},
		link: function link(scope, elem, attrs) {
			var id = scope.id || (elem[0].href.indexOf('#') > -1 ? elem[0].href.split('#')[1] : '');
			var activeClass = scope.activeClass || 'c-activate-on-scroll--active';

			Logger.debug('c-activate-on-scroll', 'linked', activeClass, id);

			if (!id) return;

			var onScroll = function onScroll() {
				var destination = document.getElementById(id);
				if (!destination) {
					return;
				}

				var offsetTop = destination.offsetTop,
				    offsetHeight = destination.offsetHeight,
				    offsetBottom = offsetTop + offsetHeight;

				var el = document.scrollingElement || document.documentElement;
				var scrollTop = el.scrollTop;
				var windowHeight = window.innerHeight;
				var scrollBottom = scrollTop + windowHeight;

				var halfOfWindowVisible = scrollBottom - windowHeight * .5;

				Logger.debug('activate-on-scroll info', id, destination, offsetTop, offsetBottom, scrollTop, halfOfWindowVisible);

				if (halfOfWindowVisible > offsetTop && halfOfWindowVisible < offsetBottom) {
					elem.addClass(activeClass);
				} else {
					elem.removeClass(activeClass);
				}
			};

			angular.element($window).bind('scroll', onScroll);
			angular.element($window).bind('load', onScroll);
		}
	};
}]);
angular.module('cevo.util').directive('cAnimateLoad', function () {
	return {
		restrict: 'A',
		link: function link(scope, elem, attrs) {
			elem.addClass('c-animate-load');
			elem.addClass('ng-hide');
			elem.on('load', function () {
				elem.addClass('ng-enter').removeClass('ng-hide');
			});

			scope.changeSrc = function () {
				elem.addClass('ng-hide').removeClass('ng-enter');
			};

			attrs.$observe('src', function () {
				scope.changeSrc();
			});
		}
	};
});
angular.module('cevo.util').directive('cAnimateView', ["$compile", function ($compile) {
	return {
		restrict: 'A',
		terminal: true,
		priority: 1000,
		link: function link(scope, element, attrs) {
			var animation = attrs.cAnimateView || 'fadeIn';
			element.attr('in-view', 'defineInViewAnimation($inview, $inviewInfo, "' + animation + '");');
			element.removeAttr('c-animate-view');
			element.addClass('c-animate-view animate-' + animation);

			scope.defineInViewAnimation = function ($inview, $inviewInfo, animation) {
				if ($inview) $inviewInfo.element.addClass('ng-enter').removeClass('ng-hide');
			};

			$compile(element)(scope);
		}
	};
}]);
angular.module('cevo.util').directive('compileHtml', ["$compile", function ($compile) {
    return {
        restrict: 'A',
        link: function link(scope, element, attrs) {
            scope.$watch(function () {
                return scope.$eval(attrs.compileHtml);
            }, function (value) {
                element.html(value);
                $compile(element.contents())(scope);
            });
        }
    };
}]);
angular.module('cevo.util').directive('cCreditCard', ["$rootScope", "Notification", "$timeout", "$filter", "Logger", function ($rootScope, Notification, $timeout, $filter, Logger) {
	return {
		restrict: 'E',
		template: '\n\t\t\t<div class="c-credit-card">\n\t\t\t\t<div class="credit-card-wrapper flex flex-center grow-between">\n\t\t\t\t\t<span class="credit-card-numbers flex">\n\t\t\t\t\t\t<input type="text" ng-model="parts[0]" maxlength=4 tabindex="{{tabindex ? firstTab : \'\'}}" />\n\t\t\t\t\t\t<input type="text" ng-model="parts[1]" maxlength=4 tabindex="{{tabindex ? secondTab : \'\'}}" />\n\t\t\t\t\t\t<input type="text" ng-model="parts[2]" maxlength=4 tabindex="{{tabindex ? thirdTab : \'\'}}" />\n\t\t\t\t\t\t<input type="text" ng-model="parts[3]" maxlength=4 tabindex="{{tabindex ? fourthTab : \'\'}}" />\n\t\t\t\t\t</span>\n\t\t\t\t\t<span class="credit-card-image">\n\t\t\t\t\t\t<i ng-if="type == \'discover\'" class="payment-card-image discover"></i>\n\t\t\t\t\t\t<i ng-if="type == \'amex\'" class="payment-card-image amex"></i>\n\t\t\t\t\t\t<i ng-if="type == \'visa\'" class="payment-card-image visa"></i>\n\t\t\t\t\t\t<i ng-if="type == \'mastercard\'" class="payment-card-image mastercard"></i>\n\t\t\t\t\t\t<span ng-if="type == \'\'" class="dim">&#8212;</span>\n\t\t\t\t\t</span>\n\t\t\t\t</div>\n\t\t\t</div>\n\t\t',
		replace: true,
		scope: {
			'number': '=',
			'type': '=',
			'tabindex': '@starttabindex'
		},
		link: function link(scope, elem, attrs) {

			scope.firstTab = scope.tabindex;
			scope.secondTab = parseInt(scope.firstTab) + 1;
			scope.thirdTab = parseInt(scope.firstTab) + 2;
			scope.fourthTab = parseInt(scope.firstTab) + 4;

			$(elem).on('click', function () {
				$(elem).find('input').each(function (i, e) {
					//$log.log('saw e with value', e.value, e.value.length);
					if (e.value.length == 0) {
						e.focus();
						return false;
					}
				});
				//elem.find('input:nth-child(1)').focus();
			});

			$(elem).find('input').on('keydown', function (ev) {
				var char = ev.which || ev.charCode;
				//$log.log('heard input keypress for credit card ', char, ev, this);

				if (this.value.length == 4 && char != 8) {
					Logger.debug('value is full, will focus next input');
					$(this).next().focus();
				}

				if (char != 8) //backspace
					return;

				if (this.value.length <= 0) {
					Logger.debug('value is empty, will focus previous input');
					$(this).prev().focus();
				}
			});

			$(elem).find('input').on('click', function () {
				if (this.value.length > 0) return false;
			});

			scope.parts = [];
			scope.$watch('parts', function () {
				scope.number = scope.parts.join('');
				var one = parseInt(scope.number.substring(0, 1));
				var two = parseInt(scope.number.substring(0, 2));
				var three = parseInt(scope.number.substring(0, 3));
				var four = parseInt(scope.number.substring(0, 4));
				var six = parseInt(scope.number.substring(0, 6));

				if (one == 4) scope.type = 'visa';else if (two == 34 || two == 37) scope.type = 'amex';else if (two >= 51 && two <= 55) scope.type = 'mastercard';else if (six >= 622126 && six <= 622925 || two == 65 || four == 6011 || three >= 644 && three <= 649) scope.type = 'discover';else scope.type = '';

				Logger.debug('creditcard saw', scope.type, scope.number, one, two, three, four, six);
			}, true);
		}
	};
}]);
angular.module('cevo.util').directive('cDragDrop', ["$q", "Logger", function ($q, Logger) {
	return {
		restrict: 'A',
		scope: {
			'file': '=cDragDrop',
			'progress': '=?cDragDropProgress',
			'type': '@cDragDropType'
		},
		link: function link(scope, elem, attrs) {
			if (typeof $ === 'undefined') {
				Logger.error('DRAGDROP', "ERROR", 'jQuery must be installed to use drag and drop.');
				return;
			}

			scope.progress = 0;

			$('body').on('dragover dragenter', function (e) {
				//Logger.debug('Heard dragover dragenter on body', e);
				$('body').addClass(e.type);
			});

			$('body').on('dragleave', function (e) {
				$('body').removeClass('dragover dragenter');
			});

			elem.addClass('c-drop-zone');

			$(elem[0]).on('dragover dragenter dragleave', function (e) {
				//Logger.debug('dragdrop', 'Saw event', e);

				if (e.type === 'dragover') elem.addClass(e.type).addClass('c-drag-hover');else if (e.type === 'dragleave') elem.removeClass('c-drag-hover dragover dragenter c-drag-enter');else if (e.type === 'dragenter') elem.addClass('c-drag-enter');

				e.preventDefault();
				e.stopPropagation();
				return false;
			});

			$(elem[0]).on('drop', function (e) {
				e.stopPropagation();
				e.preventDefault();

				elem.removeClass('c-drag-hover dragover dragenter c-drag-enter');

				var files = e.originalEvent.dataTransfer.files;
				Logger.debug('dragdrop', 'Saw dropped files', files);

				if (files.length) {
					readFile(files[0]).then(function (value) {
						Logger.debug('dragdrop', 'Read file with value', value);
						scope.file.lastModified = files[0].lastModified;
						scope.file.lastModifiedDate = files[0].lastModifiedDate;
						scope.file.name = files[0].name;
						scope.file.size = files[0].size;
						scope.file.type = files[0].type;
						scope.file.src = value;
					});
				} else {
					var url = e.originalEvent.dataTransfer.getData('url');
					Logger.debug('dragdrop', 'Read file with url', url);
					scope.file.src = url;
				}

				scope.$apply();

				return false;
			});

			function readFile(file) {
				Logger.debug('dragdrop', 'Reading file', file);
				var deferred = $q.defer();

				var reader = new FileReader();

				reader.onload = function (e) {
					deferred.resolve(e.target.result);
				};

				reader.onprogress = function (data) {
					Logger.debug('dragdrop', 'onprogress', data);
					scope.data = data;
					scope.progress = data.loaded / data.total;
				};

				reader.onerror = function (e) {
					Logger.debug('dragdrop', 'Reader onerror', e);
					deferred.reject(e);
				};

				if (scope.type == 'text') {
					Logger.debug('dragdrop', 'file reading as text ..');
					reader.readAsText(file);
				} else if (scope.type == 'buffer') {
					reader.readAsArrayBuffer(file);
				} else if (scope.type == 'binary') {
					reader.readAsBinaryString(file);
				} else {
					reader.readAsDataURL(file);
				}

				Logger.debug('dragdrop', 'Reader data url, reader', reader);

				return deferred.promise;
			}
		}
	};
}]);
angular.module('cevo.util').directive('ngEnter', function () {
	return function (scope, element, attrs) {
		element.bind('keypress', function (event) {
			if (event.which === 13) {
				if (typeof attrs.ngEnterClear !== 'undefined') this.value = '';

				scope.$apply(function () {
					scope.$eval(attrs.ngEnter);
				});

				event.preventDefault();
			}
		});
	};
});
angular.module('cevo.util').directive('cFocusOn', ["$timeout", "Logger", function ($timeout, Logger) {
	return {
		restrict: 'A',
		scope: {
			focusValue: "=cFocusOn",
			focusDelay: "@",
			blurDelay: "@"
		},
		link: function link($scope, $element, attrs) {
			$scope.$watch("focusValue", function (currentValue, previousValue) {
				if (currentValue === true) $timeout(function () {
					// $element[0].focus(); 
					var len = $element.val().length;
					$element[0].focus();
					$element[0].setSelectionRange(len, len);

					Logger.debug('focuson', $element.val(), $scope.focusDelay, $element[0]);
				}, $scope.focusDelay ? parseInt($scope.focusDelay) : 0);else if (currentValue === false && previousValue) $timeout(function () {
					$element[0].blur();
				}, $scope.blurDelay ? parseInt($scope.blurDelay) : 0);
			});
		}
	};
}]);
angular.module('cevo.util').directive('cForm', ["Logger", "Form", "$timeout", "FormValidation", function (Logger, Form, $timeout, FormValidation) {
	return {
		restrict: 'A',
		require: 'form',
		link: function link(scope, elem, attr, $form) {
			Logger.debug('Linking form', attr.name, attr.cForm);

			var watchPromise = attr.cForm || null;

			var formSubmitted = function formSubmitted($form, promise) {

				Logger.debug('form', 'heard formsubmitted from directive', promise);

				if (!promise || !promise.then) return;

				FormValidation.reset($form);

				promise.then(null, function (reason) {
					Logger.debug('form', 'heard promise error fire', reason);
					FormValidation.applyErrors($form, reason);
				});
			};

			if (watchPromise !== null) scope.$watch(watchPromise, formSubmitted.bind(null, $form));
		}
	};
}]);
angular.module('cevo.util').directive('cFormErrors', ["Logger", "Form", "$timeout", "FormValidation", function (Logger, Form, $timeout, FormValidation) {
	return {
		restrict: 'E',
		template: '\n\t\t\t<div class="c-form-errors">\n\t\t\t\t<div class="c-form-error" ng-repeat="error in element.$errors"><i class="ico ico-exclamation-circle"></i><span>{{error.message}}</span></div>\n\t\t\t</div>\n\t\t',
		replace: true,
		scope: {
			element: '='
		}
	};
}]);
angular.module('cevo.util').directive('cInfiniteScroll', ["$window", "$document", "$rootScope", "$timeout", "$compile", "Logger", function ($window, $document, $rootScope, $timeout, $compile, Logger) {
	return {
		scope: {
			infiniteScroll: '=cInfiniteScroll', //if a 'string' is passed, it will be emitted every infiniteScrollDistance (5px, 5%, etc...)...
			infiniteScrollDistance: '@cInfiniteScrollDistance', //how much distance (in px or %) scroll is from the bottom within itself, or if relative, to the window
			infiniteScrollRelative: '@cInfiniteScrollRelative', //whether or not infiniteScrollDistance is relative to the window's scroll position (true), or the element's scroll position (false)
			infiniteScrollLimit: '@cInfiniteScrollLimit', //how many infinite scroll calls should be allowed before a reset is needed
			infiniteScrollReverse: '@cInfiniteScrollReverse', //if set we will listen for scroll distance towards the top of the element / window, this is default to the bottom
			infiniteScrollDisabled: '=?cInfiniteScrollDisabled'
		},
		link: function link(scope, elem, attrs) {
			var win = angular.element($window);
			var elementHeight = elem[0].offsetHeight;
			var lastVal;

			scope.calling = false;
			var style = $window.getComputedStyle(elem[0]);
			var overflow = style['overflow'] || style['overflowY'];
			var isOverflow = overflow == 'scroll' || overflow == 'auto';
			Logger.debug('infinite scroll link overflow', elem[0].className, overflow, style.overflow, style.overflowY, isOverflow);

			var infiniteScrollReverse = typeof scope.infiniteScrollReverse !== 'undefined';

			//make sure we set this to a var because the scope.infiniteScrollRelative will keep getting reset
			if (typeof scope.infiniteScrollRelative !== 'undefined' && scope.infiniteScrollRelative !== 'false') var infiniteScrollRelative = true;else var infiniteScrollRelative = false;

			var scrollIsFunction = angular.isFunction(scope.infiniteScroll);

			if (scrollIsFunction && typeof scope.infiniteScrollLimit !== 'undefined') scope.currentLimit = parseInt(scope.infiniteScrollLimit);else scope.currentLimit = 999999999;

			Logger.debug('infinite scroll link', elem[0].className, scrollIsFunction, isOverflow, scope.infiniteScrollRelative, infiniteScrollRelative, scope.infiniteScrollDistance, scope.currentLimit);

			if (!elem.hasClass('c-infinite-scroll')) elem.addClass('c-infinite-scroll');

			var onScroll = function onScroll() {
				var element = {
					bind: elem[0].offsetHeight < elem[0].scrollHeight,

					scrollHeight: elem[0].scrollHeight,
					scrollTop: elem[0].scrollTop,
					scrollPct: elem[0].scrollTop / (elem[0].scrollHeight - elem[0].offsetHeight) * 100,
					scrollBot: elem[0].offsetHeight + elem[0].scrollTop,

					height: elem[0].offsetHeight,
					topPos: elem[0].offsetTop,
					bottomPos: elem[0].offsetTop + elem[0].offsetHeight
				};

				element.scrollRemaining = element.scrollHeight - element.scrollBot;
				element.scrollPctRemaining = element.scrollRemaining / (element.scrollHeight - element.height) * 100;

				var wdw = {
					scrollPos: $window.scrollY,
					scrollHeight: $document[0].documentElement.offsetHeight,

					height: $document[0].documentElement.clientHeight,
					bottomPosition: $window.scrollY + $document[0].documentElement.clientHeight
				};

				wdw.scrollRemaining = wdw.scrollHeight - (wdw.scrollPos + wdw.height);
				wdw.scrollPctRemaining = wdw.scrollRemaining / (wdw.scrollHeight - wdw.height) * 100;

				//window to element relative positioning
				wdw.toElementBottomPos = element.bottomPos - wdw.bottomPosition;
				wdw.toElementBottomPct = 100 - wdw.toElementBottomPos / element.height * 100;
				wdw.toElementBottomPctRemaining = wdw.toElementBottomPos / element.height * 100;

				//Logger.debug('infinite scroll', wdw, element);

				// return if limit is completed;
				if (!element.height) return;

				if (scrollIsFunction) {
					if (scope.calling || scope.currentLimit === 0 || scope.infiniteScrollDisabled) return;

					var shouldCall = false;
					if (scope.infiniteScrollDistance) {
						if (infiniteScrollReverse && !infiniteScrollRelative) {
							if (scope.infiniteScrollDistance.indexOf('%') > -1) {
								//%-based decision making.  //if undefined, we check the element, if relative, check relative to the window positioning
								//Logger.debug('infinite scroll', 'checking % up')
								shouldCall = parseInt(scope.infiniteScrollDistance.replace('%', '')) >= element.scrollPct;
							} else {
								//px-based decision making 
								shouldCall = parseInt(scope.infiniteScrollDistance.replace('%', '')) >= element.scrollTop;
							}
						} else {
							if (scope.infiniteScrollDistance.indexOf('%') > -1) {
								//%-based decision making.  //if undefined, we check the element, if relative, check relative to the window positioning
								//Logger.debug('infinite scroll', 'checking %')
								shouldCall = parseInt(scope.infiniteScrollDistance.replace('%', '')) >= (!infiniteScrollRelative ? element.scrollPctRemaining : wdw.toElementBottomPctRemaining);
							} else {
								//px-based decision making 
								//Logger.debug('infinite scroll', 'checking num', parseInt(scope.infiniteScrollDistance.replace('%', '')), ' >= ', (!infiniteScrollRelative ? 'ele '+element.scrollRemaining : 'wdw '+wdw.toElementBottomPos), infiniteScrollRelative, element.scrollRemaining, wdw.toElementBottomPos);	
								shouldCall = parseInt(scope.infiniteScrollDistance.replace('%', '')) >= (!infiniteScrollRelative ? element.scrollRemaining : wdw.toElementBottomPos);
							}
						}
					}

					if (!shouldCall) return;

					Logger.debug('infinite scroll', elem[0].className, 'calling', parseInt(scope.infiniteScrollDistance.replace('%', '')), ' >= ', !infiniteScrollRelative ? 'ele ' + element.scrollRemaining : 'wdw ' + wdw.toElementBottomPos, infiniteScrollRelative, element.scrollRemaining, wdw.toElementBottomPos);

					scope.calling = true;
					var call = scope.infiniteScroll();
					if (call && call.finally) {
						return call.finally(function () {
							scope.calling = false;
							if (scope.currentLimit > 0) scope.currentLimit--;
						});
					} else {
						scope.calling = false;
					}
				} else if (scope.infiniteScroll) {
					if (typeof scope.infiniteScrollDistance === 'undefined') scope.infiniteScrollDistance = '3%'; //default 5%

					var shouldCall = false;

					if (scope.infiniteScrollDistance.indexOf('%') > -1) var curVal = infiniteScrollRelative ? wdw.scrollPctRemaining : element.scrollPctRemaining;else var curVal = infiniteScrollRelative ? wdw.scrollRemaining : element.scrollRemaining;

					var bounce = parseInt(scope.infiniteScrollDistance.replace('%', '').replace('px', ''));

					if (typeof lastVal === 'undefined' || lastVal < curVal) lastVal = curVal;

					if (lastVal - curVal >= bounce) shouldCall = true;

					//Logger.debug('infinite scroll', scope.infiniteScroll, infiniteScrollRelative, shouldCall, lastVal - curVal, bounce, scope.infiniteScrollDistance);

					if (shouldCall) {
						$rootScope.$emit(scope.infiniteScroll, { window: wdw, element: element });
						lastVal = curVal;
						if (scope.currentLimit > 0) scope.currentLimit--;
					}
				}
			};

			// set initial data and limit
			scope.initialLimit = function () {
				if (scope.currentLimit === 0) {
					scope.currentLimit = scope.infiniteScrollLimit;
					onScroll();
				}
			};

			elem.bind('scroll', onScroll);
			win.bind('scroll', onScroll);

			scope.$on('$destroy', function () {
				//Logger.debug('infinite scroll unbind', elem);
				elem.unbind('scroll', onScroll);
				win.bind('scroll', onScroll);
			});

			// Check on next event loop to give the browser a moment to load more.
			//$timeout(onScroll);
		},

		controller: ["$scope", function controller($scope) {
			//Logger.debug('infinite', 'controller', $scope);
			$scope.$watch("currentLimit", function (newValue, oldValue) {
				//Logger.debug('infinite', 'current limit', $scope.currentLimit);
				$scope.currentLimit = newValue;
			});

			this.getScope = function () {
				return $scope;
			};

			this.getLimit = function () {
				return $scope.currentLimit;
			};

			this.resetLimit = function () {
				$scope.initialLimit();
			};
		}]
	};
}]);

angular.module('cevo.util').directive('cInfiniteScrollLimitReset', ["$window", "$rootScope", "Logger", function ($window, $rootScope, Logger) {
	return {
		scope: true,
		require: "^cInfiniteScroll",
		link: function link(scope, elem, attrs, ctrlInfinite) {
			// set hide button to load more
			var parentScope = ctrlInfinite.getScope();
			scope.loadButton = false;

			parentScope.$watch('currentLimit', function (newValue, oldValue) {
				if (newValue === 0) scope.loadButton = true;
			});

			scope.resetLimit = function () {
				if (ctrlInfinite.getLimit() === 0) {
					scope.loadButton = false;
					ctrlInfinite.resetLimit();
				} else {
					scope.loadButton = true;
				}
			};
		}
	};
}]);
angular.module('cevo.util').directive('cInput', ["Logger", function (Logger) {
	return {
		restrict: 'E',
		template: '\n\t\t\t<div class="c-input animated" ng-class="status == \'inactive\' ? \'dim\' : status" ng-style="{ \'width\' : (width && status == \'active\' ? width : \'auto\')}" ng-click="status = \'active\'">\n\t\t\t\t<span class="label">\n\t\t\t\t\t<i class="ico" ng-class="ico"></i>\n\t\t\t\t\t<span ng-bind-html="label">{{label}}</span>\n\t\t\t\t</span>\n\t\t\t\t<input placeholder="{{label}}" size="size" ng-model="model" ng-blur="blurInput()" />\n\t\t\t</div>',
		replace: true,
		scope: {
			'ico': '@',
			'label': '@',
			'model': '=',
			'width': '@',
			'size': '@'
		},
		link: function link(scope, elem, attrs) {
			scope.status = 'inactive';
			var originalSize = elem[0].offsetWidth;

			scope.$watch('status', function () {
				if (scope.status == 'active') elem[0].getElementsByTagName('input')[0].focus();else resetWidth();
			});

			function resetWidth() {
				//we can't transition from auto=>defined width (css3 issue), so we need to set the width to a static value
				elem[0].style.width = originalSize;
			}

			scope.blurInput = function () {
				if (!scope.model) scope.status = 'inactive';
			};
		}
	};
}]);
angular.module('cevo.util').directive('inputNumeric', function () {
	return {
		require: 'ngModel',
		link: function link(scope, element, attr, ngModelCtrl) {
			function fromUser(text) {
				var transformedInput = text.replace(/[^0-9]/g, '');

				if (transformedInput !== text) {
					ngModelCtrl.$setViewValue(transformedInput);
					ngModelCtrl.$render();
				}

				return transformedInput; // or return Number(transformedInput)
			}

			ngModelCtrl.$parsers.push(fromUser);
		}
	};
});
angular.module('cevo.util').directive('modal', ["$rootScope", "$timeout", "$interval", "$sce", "Logger", function ($rootScope, $timeout, $interval, $sce, Logger) {
	return {
		restrict: 'E',
		template: '\n\t\t\t<div class=\'ng-modal\' ng-show=\'show\'>\n\t\t\t\t<div class=\'ng-modal-overlay\' ng-if="show" ng-click=\'hideModal("overlay")\'></div>\n\t\t\t\t<div class=\'ng-modal-dialog\' ng-if="show">\n\t\t\t\t\t<a class=\'ng-modal-close ico ico-times\' href="" ng-if="!settings.hideCancel || !settings.hideCornerCancel" ng-click=\'hideModal(); cancelClick();\'></a>\n\t\t\t\t\t<p class="ng-modal-title" ng-show="title">\n\t\t\t\t\t\t<span ng-bind-html="title" class=""></span>\n\t\t\t\t\t</p>\n\t\t\t\t\t<div class=\'ng-modal-dialog-content\'>\n\t\t\t\t\t\t<div ng-transclude></div>\n\t\t\t\t\t</div>\n\t\t\t\t</div>\n\t\t\t</div>',
		scope: {
			show: '=',
			hideCancel: '@hidecancel',
			cancelAction: '&?',
			hideCornerCancel: '@hidecornercancel',
			hideButtonCancel: '@hidebuttoncancel',
			cancelText: '@canceltext',
			cancelClick: '&cancelclick'
		},
		replace: true, // Replace with the template below
		transclude: true, // we want to insert custom content inside the directive
		link: function link(scope, element, attrs) {
			scope.overlayCancel = attrs.overlayCancel === 'true';

			Logger.debug('Linking modal with cancel settings', scope.hideCancel, scope.hideCornerCancel, scope.hideButtonCancel);
			scope.title = $sce.trustAsHtml(attrs.mtitle);
			Logger.debug('trusting title', scope.title, attrs.mtitle);

			scope.settings = {
				hideCancel: scope.hideCancel !== 'false',
				hideCornerCancel: scope.hideCornerCancel != 'false',
				hideButtonCancel: scope.hideButtonCancel != 'false'
			};

			scope.$watch('hideCornerCancel', function () {
				scope.settings.hideCornerCancel = scope.hideCornerCancel != 'false';
			});

			scope.$watch('show', function () {
				if (scope.show) $rootScope.$emit('modal:active');else $rootScope.$emit('modal:inactive');
			});

			scope.hideModal = function (method) {
				if (method == 'overlay' && !scope.overlayCancel) return;

				scope.show = false;

				if (scope.cancelAction) {
					//keep this here, removes any delay and immediately updates the scope
					$timeout(function () {
						scope.cancelAction();
					});
				}
			};
		}
	};
}]);
angular.module('cevo.util').directive('cNotifications', ["Notification", "Logger", function (Notification, Logger) {
	return {
		restrict: 'E',
		template: '\n\t\t\t<div class="c-notifications notifications-wrapper"">\n\t\t\t\t<div class="c-notification" ng-class="notification.style" ng-repeat="notification in notifications | filter:{ type : \'global\' }" ng-click="notifications.splice($index, 1);">\n\t\t\t\t\t<span ng-if="notification.style == \'error\'" class="ico ico-exclamation-circle"></span>\n\t\t\t\t\t<span ng-if="notification.style == \'success\'" class="ico ico-check-circle"></span>\n\t\t\t\t\t<p>{{notification.message}}</p>\n\t\t\t\t\t<span class="ico ico-times"></span>\n\t\t\t\t</div>\n\t\t\t</div>\n\t\t',
		replace: true,
		scope: {},
		link: function link(scope, elem, attrs) {
			Logger.debug('Attaching listener to Notification from cNotifications');
			Notification.listen(function () {
				Logger.debug('Heard notification listener trigger in Notifications', Notification.notifications);
				scope.notifications = Notification.notifications;
			});
		}
	};
}]);
angular.module('cevo.util').directive('cSearch', ["$timeout", "$filter", function ($timeout, $filter) {
	return {
		restrict: 'E',

		//chrome ignores autocompleted="off", so use "new-password" instead
		template: '\n\t\t\t<div class="c-search" ng-class="[active ? \'focus\' : \'\', curVal != \'\' ? \'ng-not-empty\' : \'\']">\n\t\t\t\t<div class="input-wrapper icon">\n\t\t\t\t\t<i class="ico ico-magnify"></i>\n\t\t\t\t\t<input type="text" autocomplete="new-password" ng-model-options="{ debounce: 200 }" ng-model="curVal" ng-focus="active=true" ng-blur="blur()" ng-enter="triggerEnter()" placeholder="{{placeholder}}" />\n\t\t\t\t</div>\n\t\t\t\t<div class="dropdown" ng-class="visible && results.length > 0 ? \'visible\' : \'\'">\n\t\t\t\t\t<p class="" ng-repeat="row in results | limitTo:limit track by $index" ng-click="selectValue((val == \'{}\' ? row : row[val]), row[label]);">\n\t\t\t\t\t\t<span class="">{{row[label]}}</span>\n\t\t\t\t\t\t<i ng-if="icon" class="ico" ng-class="[\'ico-\'+icon]"></i>\n\t\t\t\t\t</p>\n\t\t\t\t</div>\n\t\t\t</div>',
		replace: true,
		scope: {
			'data': '=?',
			'val': '@',
			'label': '@',
			'icon': '@',
			'title': '@',
			'model': '=?',
			'onSelect': '&',
			'onChange': '&',
			'onEnter': '&',
			'default': '=',
			'filter': '@',
			'limit': '@',
			'curVal': '=',
			'placeholder': '@'
		},
		link: function link(scope, elem, attrs) {
			scope.visible = false;
			scope.active = false;
			scope.results = [];
			scope.onSelectClear = typeof attrs.onSelectClear !== 'undefined';
			var selected = false;

			if (typeof scope.limit === 'undefined') scope.limit = 5;

			scope.$watch('data', function () {
				if (!scope.data || scope.data.length <= 0) scope.visible = false;

				scope.results = scope.data;
			}, true);

			scope.$watch('curVal', function (newVal, oldVal) {
				if (!selected && typeof oldVal !== 'undefined' && oldVal != newVal && scope.curVal && scope.curVal.length > 0) scope.visible = true;else selected = false;

				if (scope.curVal && typeof scope.onChange === 'function' && typeof scope.onChange() === 'function') scope.onChange()(scope.curVal);

				if (scope.curVal && scope.filter) {
					var filter = {};
					filter[scope.label] = scope.curVal;
					scope.results = $filter('filter')(scope.data, filter, scope.startsWith);
				}
			});

			scope.triggerEnter = function () {
				if (typeof scope.onEnter === 'function' && typeof scope.onEnter() === 'function') scope.onEnter()(scope.curVal);
			};

			elem.on('click', function () {
				scope.visible = false;
			});

			$timeout(function () {
				attrs.$set('style', 'min-width : ' + elem.offsetWidth + 'px');
			}, 300);

			scope.selectValue = function (val, label) {
				selected = true;
				scope.curVal = label;
				scope.visible = false;

				if (typeof scope.model !== 'undefined') scope.model = val;

				if (typeof scope.onSelect === 'function' && typeof scope.onSelect() === 'function') {
					scope.onSelect()(val);

					if (scope.onSelectClear) scope.curVal = '';
				}
			};

			scope.startsWith = function (actual, expected) {
				var lowercase = actual.toLowerCase();
				return lowercase.indexOf(expected.toLowerCase()) === 0;
			};

			scope.blur = function () {
				scope.active = false;

				$timeout(function () {
					scope.visible = false;
				}, 100);
			};
		}
	};
}]);
var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

angular.module('cevo.util').directive('cSelect', ["$timeout", function ($timeout) {
	return {
		restrict: 'E',

		//chrome ignores autocompleted="off", so use "new-password" instead
		template: '\n\t\t\t<div class="c-select" ng-class="[active ? \'focus\' : \'\', curVal != \'\' ? \'ng-not-empty\' : \'\']">\n\t\t\t\t<input type="text" autocomplete="new-password" ng-model="filter.text" ng-focus="active=true" c-focus-on="active" ng-blur="inactive()" />\n\t\t\t\t<p class="current-value">{{curVal ? curVal : title}}</p>\n\t\t\t\t<div class="dropdown" ng-show="active" ng-class="active ? \'visible\' : \'\'">\n\t\t\t\t\t<p ng-if="default[label]" ng-click="selectValue(default)">\n\t\t\t\t\t\t{{default[label]}}\n\t\t\t\t\t</p>\n\t\t\t\t\t<p class="" ng-repeat="row in data | filter:filter.text track by $index" ng-click="selectValue(row);">\n\t\t\t\t\t\t{{row[label]}}\n\t\t\t\t\t</p>\n\t\t\t\t</div>\n\t\t\t</div>',
		replace: true,
		scope: {
			'data': '=',
			'val': '@',
			'label': '@',
			'title': '@',
			'model': '=',
			'default': '=?'
		},
		link: function link(scope, elem, attrs) {
			scope.curVal = '';
			scope.visible = true;
			scope.active = false;
			scope.filter = {
				'text': ''
			};

			elem.on('click', function () {
				scope.$apply(function () {
					scope.active = true;
				});
			});

			scope.$watch('active', function () {
				if (!scope.active) scope.filter.text = '';
			});

			scope.$watch('model', function () {
				scope.selectModelValue();
			});

			scope.$watch('data', function () {
				scope.set = false;
				if (!scope.data || scope.data.length <= 0) return;

				scope.selectModelValue();

				if (scope.default && !scope.set) scope.selectValue(scope.default);
			}, true);

			scope.$watch('curVal', function () {});

			$timeout(function () {
				attrs.$set('style', 'min-width : ' + elem[0].offsetWidth + 'px');
			}, 300);

			scope.selectModelValue = function () {
				if (_typeof(scope.model) !== undefined) {
					for (var i in scope.data) {
						if (scope.val == '{}' && scope.data[i] == scope.model || scope.val != '{}' && scope.data[i][scope.val] == scope.model) {
							scope.selectValue(scope.data[i]);
							scope.set = true;
							break;
						}
					}
				}
			};

			scope.selectValue = function (obj) {
				if (scope.val == '{}' && typeof scope.model !== 'undefined' && scope.model[scope.label] != obj[scope.label] || scope.val != '{}' && scope.model != obj[scope.val]) scope.model = scope.val == '{}' ? obj : obj[scope.val];

				scope.curVal = obj[scope.label];

				scope.visible = false;
				$timeout(function () {
					scope.visible = true;
				}, 200);
			};

			scope.inactive = function () {
				$timeout(function () {
					//need to allow the click to happen before we hide the dropdown
					scope.active = false;
				}, 150);
			};
		}
	};
}]);
angular.module('cevo.util').directive('cSticky', ["$window", "$timeout", "Logger", function ($window, $timeout, Logger) {
	return {
		scope: {
			lSticky: '@',
			stickyWindow: '@', //should this be sticky to the window or to the parent element? usually parent element is fine.  default is false
			stickyWindowOffset: '@', //should be used to manually change offsetTop by the amount passed. Useful if there is more than one sticky element on a page
			stickyDisabled: '=?',
			stickyBottom: '@',
			stickyDebug: '@'
		},
		link: function link(scope, element, attrs) {
			var $win = angular.element($window);
			var stickyClass = scope.sticky || 'c-sticky',
			    offsetTop,
			    offsetHeight,
			    offsetWidth,
			    offsetBottom,
			    position,
			    bodyContainer,
			    stickyContainer,
			    stickyWindow,
			    stickyBottom,
			    stickyDebug,
			    scrollTop,
			    scrollHeight,
			    stickyContainerHeight,
			    elementPosition,
			    windowOffset,
			    windowHeight,
			    windowBottom,
			    elementShouldBottom;

			Logger.debug('sticky', 'sticky from util');

			stickyDebug = typeof scope.stickyDebug !== 'undefined' && scope.stickyDebug !== 'false';
			element.addClass(stickyClass);

			if (stickyDebug) Logger.debug('sticky', 'sticky settings', stickyClass, offsetTop, '=', element[0].offsetTop, 'height', offsetHeight, 'width', offsetWidth, stickyContainer, stickyWindow, scope.stickyWindow);

			var setDefaults = function setDefaults() {
				position = element[0].style.position;
				offsetTop = element[0].offsetTop;
				offsetHeight = element[0].offsetHeight;
				offsetWidth = element[0].offsetWidth;
				offsetBottom = offsetTop + offsetHeight;
				stickyContainer = element.parent();

				bodyContainer = document.body;

				if (scope.stickyWindowOffset) offsetTop -= parseInt(scope.stickyWindowOffset);

				if (stickyDebug) Logger.debug('sticky', 'set defaults', position, offsetHeight, offsetWidth, offsetTop, offsetBottom, windowOffset, windowHeight, scrollTop);
			};

			var onScroll = function onScroll(event) {
				if (scope.stickyDisabled) return;

				position = element[0].style.position;
				stickyWindow = typeof scope.stickyWindow !== 'undefined' && scope.stickyWindow !== 'false';
				stickyBottom = typeof scope.stickyBottom !== 'undefined' && scope.stickyBottom !== 'false';
				stickyDebug = typeof scope.stickyDebug !== 'undefined' && scope.stickyDebug !== 'false';

				var el = document.scrollingElement || document.documentElement;
				//if (stickyDebug)
				//	Logger.debug('sticky', 'document element', el);

				scrollTop = el.scrollTop;
				scrollHeight = el.scrollHeight;
				stickyContainerHeight = stickyContainer[0].offsetHeight;
				stickyContainerOffset = stickyContainer[0].offsetTop;
				elementPosition = scrollTop - offsetTop + offsetHeight - stickyContainerHeight;
				//elementPositionFromContainerBottom = 
				//elementPositionFromContainerBottom;

				windowOffset = window.pageYOffset;
				windowHeight = window.innerHeight;
				windowBottom = windowOffset + windowHeight;
				elementShouldBottom = windowBottom > offsetBottom;

				if (stickyDebug) Logger.debug('sticky', 'element position', position, offsetHeight, offsetWidth, offsetTop, offsetBottom, windowOffset, windowHeight, scrollTop, elementPosition);

				//Logger.debug('sticky', element[0].style.position, scrollTop, offsetTop, offsetHeight, stickyContainer.offsetHeight, scrollTop - offsetTop + element[0].offsetHeight - stickyContainer.offsetHeight);

				//instead of checking for a class, let's just check to see if the position is "fixed"
				//activate the fixed
				if ((scrollTop >= offsetTop || stickyBottom && elementShouldBottom) && element[0].style.position != 'fixed') {
					setDefaults();
					if (stickyDebug) Logger.debug('sticky', 'stickying nav ...', position, offsetHeight, offsetWidth, offsetTop, offsetBottom, windowOffset, windowHeight, scrollTop);

					element.removeClass(stickyClass + '-inactive').addClass(stickyClass + '-active');

					if (!stickyBottom) stickyContainer.css('height', stickyContainerHeight + 'px'); //set the height of the container so we don't get weird snapping behavior

					element.css('position', 'fixed');
					element.css('top', stickyBottom ? 'auto' : scope.stickyWindowOffset ? parseInt(scope.stickyWindowOffset) + 'px' : 0);
					element.css('bottom', stickyBottom ? 0 : scope.stickyWindowOffset ? parseInt(scope.stickyWindowOffset) + 'px' : 'auto');
					element.css('left', 'auto');
					element.css("width", offsetWidth + "px");

					if (offsetHeight) element.css('height', offsetHeight);else element.css('height', 'auto');

					//element.css('z-index', 2);  do this in CSS, don't hardcode here
				} else if (element[0].style.position == 'fixed' && (!stickyBottom && scrollTop <= offsetTop || stickyBottom && (windowBottom < offsetBottom || scrollTop == 0))) {
					if (stickyDebug) Logger.debug('sticky', 'unstickying nav ...', position, offsetHeight, offsetWidth, offsetTop, offsetBottom, windowOffset, windowHeight, scrollTop, !stickyBottom && scrollTop <= offsetTop, windowBottom < offsetBottom, stickyBottom, scrollTop == 0);

					element.removeClass(stickyClass + '-active').addClass(stickyClass + '-inactive');
					stickyContainer.css('height', 'auto');

					element.css('position', 'static');
					element.css("width", "auto");
					element.css("top", "auto");
					element.css("bottom", "auto");
					//reset some element settings in case the viewport changes
					setDefaults();
				} else if (element[0].style.position == 'fixed' && elementPosition > 0 && !stickyWindow) {
					//make it appear to stay in position, hug bottom of parent
					//element position is relative to the container element, so it will be > 0 (typically it measures distance from bottom of container element)
					var origin = scope.stickyWindowOffset || 0;

					if (stickyDebug) {
						Logger.debug('sticky', 'setting inverse scroll top position ...', origin - elementPosition);
					}

					//if element top < 0, 
					element.css("top", origin - elementPosition + "px");
				} else if (element[0].style.position != 'fixed') {//if not fixed and not scrolling, reset defaults to make sure it's up to date.  This helps solve issues with dynamic content.
					//if (stickyDebug)
					//Logger.debug('sticky', 'setting defaults on scroll ...');

					//setDefaults();
				}
			};

			setDefaults();

			$win.bind('scroll', onScroll);

			scope.$on('$destroy', function () {
				$win.unbind('scroll', onScroll);
			});
		}
	};
}]);
angular.module('cevo.util').directive('cTable', ["$filter", "$timeout", "$sce", "$compile", "$state", "cevo", "Logger", function ($filter, $timeout, $sce, $compile, $state, cevo, Logger) {
	return {
		restrict: 'E',
		template: '\n\t\t\t<div class="c-table-wrapper clearfix" ng-class="[loading ? \'loading\' : \'\']">\n\t\t\t\t<div class="c-table-container">\n\t\t\t\t\t<table class="c-table classic" ng-class="[settings.sortable ? \'sortable\' : \'\', \'sorted-by-\'+(orderBy.replace(\'.val\', \'\')), \'dir-\'+(direction == \'-\' ? \'desc\' : \'asc\')]">\n\t\t\t\t\t\t<colgroup ng-repeat="(i, col) in columns track by $index"></colgroup>\n\t\t\t\t\t\t<thead ng-repeat-start="(groupIndex, group) in (groups ? groups : [0]) track by $index">\n\t\t\t\t\t\t\t<tr class="">\n\t\t\t\t\t\t\t\t<th ng-class="[(val.align ? \'align-\'+val.align : \'\'), (!settings.sortable && val.sortable !== \'true\' ? \'nosort\' : \'\')]" ng-click="(settings.sortable || val.sortable === \'true\' ? orderRows(val.key || val, val) : \'\')" ng-repeat="(key, val) in columns track by $index">\n\t\t\t\t\t\t\t\t\t<span class="" ng-class="val.align == \'center\' || !val.align ? \'grow-out\' : \'\'">\n\t\t\t\t\t\t\t\t\t\t<span>{{groups && group[key] ? group[key] : key}}</span>\n\t\t\t\t\t\t\t\t\t\t<i class="ico" ng-if="orderBy.replace(\'.val\', \'\') == (val.key ? val.key : val)" ng-class="direction == \'-\' ? \'ico-caret-down\' : \'ico-caret-up\'"></i>\n\t\t\t\t\t\t\t\t\t</span>\n\t\t\t\t\t\t\t\t</th>\n\t\t\t\t\t\t\t</tr>\n\t\t\t\t\t\t</thead>\n\t\t\t\t\t\t<tbody ng-repeat-end>\n\t\t\t\t\t\t\t<tr ng-click="redirect(tr)" ng-repeat="tr in tbody[groupIndex] | orderBy:orderBy:direction | limitTo:settings.pageSize:settings.index track by $index">\n\t\t\t\t\t\t\t\t<td ng-class="[(val.key || val).toLowerCase().replace(\' \', \'-\'), tr[val.key || val].class || \'\', (val.align ? \'align-\'+val.align : \'\')]" ng-repeat="(key,val) in columns track by $index">\n\t\t\t\t\t\t\t\t\t<span ng-if="!tr[val.key || val].type" compile-html="((val.key ? val.key : val) == \'$index\' ? ($parent.$parent.$index+1+((settings.page-1)*settings.pageSize)) : tr[val.key || val].val)"></span>\n\t\t\t\t\t\t\t\t\t<span ng-if="tr[val.key || val].type == \'link\'"><a ui-state="tr[val.key || val].sref" ui-state-params="tr[val.key || val].params" ng-bind-html="tr[val.key || val].val ? tr[val.key || val].val : tr[val.key || val].label"></a></span>\n\t\t\t\t\t\t\t\t\t<span ng-if="tr[val.key || val].type == \'button\'"><a ui-state="tr[val.key || val].sref" ui-state-params="tr[val.key || val].params" ng-bind-html="tr[val.key || val].val ? tr[val.key || val].val : tr[val.key || val].label" class="button outline square"></a></span>\n\t\t\t\t\t\t\t\t\t<span ng-if="tr[val.key || val].type == \'date\' && tr[val.key || val].val > 0 && tr[val.key || val].val < 99999999999">{{tr[val.key || val].val*1000 | amDateFormat:(tr[val.key || val].dateFormat ? tr[val.key || val].dateFormat : \'ddd, MMM Do h:mm A\')}} <span ng-if="tr[val.key || val].timezone">{{tr[val.key || val].val*1000 | timezoneAbbreviation:false}}</span></span>\n\t\t\t\t\t\t\t\t</td>\n\t\t\t\t\t\t\t</tr>\n\t\t\t\t\t\t</tbody>\n\t\t\t\t\t</table>\n\t\t\t\t\t<div ng-if="showContext" id="context-wrapper" ng-style="{ \'min-height\' : lastHeight+\'px\' }">\n\t\t\t\t\t\t<div ng-if="isLoading"><span class="message" ng-bind-html="safeLoadingMessage || \'Loading data ...\'"></span></div>\n\t\t\t\t\t\t<div ng-if="noData && !isLoading" class="no-data"><span class="dim" ng-bind-html="safeNoDataMessage || \'There is no data to show\'"></span></div>\n\t\t\t\t\t</div>\n\t\t\t\t</div>\n\t\t\t\t<div class="pagination-wrapper" ng-if="settings.pagination && settings.numPages > 0">\n\t\t\t\t\t<div class="pagination">\n\t\t\t\t\t\t<a href="" class="" ng-class="settings.page <= 1 ? \'inactive\' : \'\'" ng-click="decrementPage()"><i class="ico ico-angle-left"></i></a>\n\t\t\t\t\t\t<span class="page-list">\n\t\t\t\t\t\t\t<a href="" class="" ng-click="settings.page = (i+1)" ng-class="settings.page == (i+1) ? \'active\' : \'\'" ng-repeat="(i, val) in (\'.\').repeat(settings.numPages) track by $index">\n\t\t\t\t\t\t\t\t<span>{{i+1}}</span>\n\t\t\t\t\t\t\t</a>\n\t\t\t\t\t\t</span>\n\t\t\t\t\t\t<a href="" class="" ng-class="settings.page >= settings.numPages ? \'inactive\' : \'\'" ng-click="incrementPage()"><i class="ico ico-angle-right"></i></a>\n\t\t\t\t\t</div>\n\t\t\t\t\t<p class="pagination-results">Showing {{settings.index+1}}-{{settings.index+settings.pageSize > tbody.length ? tbody.length : settings.index+settings.pageSize}} of {{tbody.length}} Results</p>\n\t\t\t\t</div>\n\t\t\t</div>',
		replace: true,
		scope: {
			'data': '=',
			'groups': '=',
			'columns': '=',
			'filters': '=',
			'page': '@',
			'pageSize': '@',
			'pagination': '@',
			'multi': '=?',
			'orderBy': '@',
			'sortable': '@',
			'direction': '@',
			'loading': '=',
			'loadingMessage': '@',
			'noDataMessage': '@',
			'srefState': '@',
			'srefKey': '@',
			'srefValue': '@'
		},
		link: function link(scope, elem, attrs) {
			//scope.page = scope.page || 1;
			//scope.pageSize = scope.pageSize || 25;
			//scope.index = 0;

			scope.safeNoDataMessage = $sce.trustAsHtml(scope.noDataMessage);
			Logger.debug('trusting no data message, making safe', scope.safeNoDataMessage);
			scope.safeLoadingMessage = $sce.trustAsHtml(scope.loadingMessage);
			Logger.debug('trusting loading message, making safe', scope.safeLoadingMessage);

			scope.mutableRows = {};
			scope.rows = {};
			scope.tbody = {};

			scope.firstLoad = true;

			scope.direction = typeof scope.direction === 'undefined' ? '-' : scope.direction;
			Logger.debug('ctable', 'linking sortable with', scope.sortable);
			Logger.debug('ctable', 'orderBy', scope.orderBy);
			scope.lastHeight;
			scope.showContext = true;
			scope.settings = {
				pagination: scope.pagination !== 'false',
				page: parseInt(scope.page) || 1,
				pageSize: parseInt(scope.pageSize) || 25,
				sortable: scope.sortable !== 'false',
				numPages: 1,
				index: 0

				/*
    if (typeof(scope.pagination) !== 'undefined')
    {
    	if (typeof(scope.dataGroups) !== 'undefined')
    		scope.settings.pagination = false;
    }
    */

			};if (!scope.settings.pagination) {
				scope.settings.pageSize = 99999999999;
				Logger.debug('ctable', 'Table pagination is off, setting page size to huge', scope.settings.pagination, scope.settings.pageSize);
			}

			Logger.debug('ctable', 'Table settings', 'pagination', scope.settings.pagination, 'pagesize', scope.settings.pageSize, 'page', scope.settings.page);

			var initialDirection = scope.direction;

			scope.incrementPage = function () {
				if (scope.settings.page < scope.settings.numPages) scope.settings.page++;
			};

			scope.decrementPage = function () {
				if (scope.settings.page > 1) scope.settings.page--;
			};

			scope.calculatePages = function (group) {
				if (typeof group === 'undefined') group = 0;

				if (scope.settings.pagination) {
					scope.settings.numPages = Math.ceil((scope.mutableRows[group] ? scope.tbody[group].length : 1) / scope.settings.pageSize);

					Logger.debug('ctable', 'Calculated num pages', scope.settings.numPages, scope.mutableRows[group] ? 'mutable rows ' + scope.tbody[group].length : 'static ' + 1, 'pagesize', scope.settings.pageSize);
				} else {
					Logger.debug('ctable', 'Pagination not enabled ', scope.settings.pagination);
					scope.settings.numPages = 1;
				}
			};

			scope.orderRows = function (key, onLoad) {
				Logger.debug('ordering rows because', scope.settings.sortable, onLoad);
				if (key === undefined || typeof key === 'undefined' || key === null || key.indexOf('.val') < 0) {
					key = key + '.val'; //we want to always sort by the "val" attribute of the object, but the user creating the table won't know to include .val
				}
				if (key == scope.orderBy) {
					if (onLoad === true) {
						scope.orderBy = key;
						scope.direction = initialDirection;
					} else {
						if (scope.direction == '-') scope.direction = '';else scope.direction = '-';
					}
				} else {
					scope.orderBy = key;
					scope.direction = initialDirection;
				}
			};

			/*
   scope.filterRows = function() {
   	for (var i in scope.groups)
   	{
   		var filters = {};
   		for (var key in scope.filters) //need to convert filter to compatable filter
   			filters[key] = { 'val' : scope.filters[key]};
   			scope.tbody[i] = $filter('filter')(scope.mutableRows[i], filters, false);
   		scope.calculatePages(i);
   	}
   }
   */

			scope.calculateHeight = function () {
				$timeout(function () {
					scope.lastHeight = elem[0].getElementsByTagName('tbody')[0].offsetHeight;
				}, 100);
			};

			scope.registerColumnHover = function () {
				var table = elem.find('table');
				//Logger.debug('ctable', table);

				var tds = angular.element(table[0]).find('td');
				var ths = angular.element(table[0]).find('th');
				//children = children.concat(angular.element(table[0]).find('td'));

				//Logger.debug('ctable', 'Found children', tds, ths, table[0]);

				tds.on('mouseover mouseout', scope.hoverColumnCallback);
				ths.on('mouseover mouseout', scope.hoverColumnCallback);
			};

			scope.hoverColumnCallback = function (e) {
				if (typeof $ === 'undefined') return;

				var table = elem.find('table');

				if (e.type == 'mouseover') {
					//Logger.debug('ctable', 'Heard', e.type, 'event', this);
					var element = table.find('colgroup').eq($(this).index());
					//Logger.debug('ctable', 'Found element', element);
					element.addClass('hover');
				} else {
					//Logger.debug('ctable', 'Heard', e.type, 'event');
					var element = table.find('colgroup').eq($(this).index());
					//Logger.debug('ctable', 'Found element', element);
					element.removeClass('hover');
				}

				e.stopPropagation();
			};

			scope.handleData = function (data, group) {
				Logger.debug('ctable', 'Handling data', data, group);

				scope.noData = false;

				scope.rows[group] = [];
				scope.mutableRows[group] = [];

				for (var i in data) {
					for (var r in data[i]) {
						if (angular.isObject(data[i][r])) {
							Logger.debug('ctable', 'Saw object, looping', r, data[i][r]);
							var obj = {};
							for (var x in data[i][r]) {
								data[i][r + '.' + x] = data[i][r][x];
							}delete data[i][r];
						}
					}

					scope.rows[group].push(data[i]); // = scope.data[i];
				}

				Logger.debug('ctable', 'Pushed data into rows', group, scope.rows[group]);

				for (var i in scope.rows[group]) {
					var row = scope.rows[group][i];
					Logger.debug('ctable', 'Processing row', row);

					var tr = {};

					for (var k in scope.columns) {
						var col = scope.columns[k];
						var td = {
							'val': col.val,
							'filter': col.filter,
							'col': k,
							'key': col.key || col,
							'type': col.type,
							'class': col.class,
							'sref': col.sref,
							'label': $sce.trustAsHtml(col.label),
							'dateFormat': col.dateFormat,
							'timezone': col.timezone || false,
							'eval': col.eval,
							'params': {}
						};

						var key = col.key || col;

						if (col.eval) //we're evaling the key isntead of pulling it from the row
							{
								//if we eval, we need to set a key to override "Object"
								if (!col.key) {
									scope.columns[k].key = k;
									td.key = k;
								}

								var $rowScope = scope.$new(); //we need to create a scope to use for evaling code
								$rowScope = angular.extend($rowScope, row); //add all row data to the scope so we can use it to eval
								td.val = $rowScope.$eval(col.eval);
								Logger.debug('Saw eval', col.eval, td.val);
							} else {
							if (key == '$index') //this will allow us to use the row index as a value
								td.val = parseInt(i) + 1;else {
								//Logger.debug('ctable', 'trying to trust', row[key]);

								if (angular.isString(row[key])) td.val = $sce.trustAsHtml(row[key]);else td.val = row[key];
							}
						}

						if (col.align) td.class += ' align-' + col.align;

						if (td.filter) {
							var parts = td.filter.split(':');
							if (parts.length > 1) {
								var filterName = parts.shift();
								td.val = $filter(filterName)(td.val, parts);
								if (filterName == 'number') td.val = parseFloat(td.val);
							}
						}

						if (col.type) //do it this way so we can support _custom1, 2, 3, etc...
							{
								//this is all so that we can parse row data to create ui-sref params
								if (col.params) {
									//loop each param
									for (var p in col.params) {
										var param = col.params[p];
										//pull value out and get it from the row
										var matches;
										while (matches = param.match(/\{\{([a-zA-Z0-9]*)\}\}/)) {
											//replace the param so the while loop doesn't infinitely loop
											param = param.replace(matches[0], row[matches[1]]);
											td.params[p] = param;
										}
									}
								}
							}

						tr[td.key] = td;
					}

					if (scope.srefState && scope.srefKey && scope.srefValue) {
						tr['sref'] = {
							'state': scope.srefState,
							'key': scope.srefKey,
							'value': row[scope.srefValue]
						};
					}

					scope.mutableRows[group].push(tr);
				}

				scope.tbody[group] = scope.mutableRows[group];
				scope.showContext = scope.tbody[group].length <= 0;

				Logger.debug('ctable', 'Finished sorting data', scope.tbody, scope.mutableRows, scope.firstLoad);

				if (scope.orderBy && scope.direction && scope.firstLoad === true) {
					scope.orderRows(scope.orderBy, scope.firstLoad);
					scope.firstLoad = false;
				}

				//scope.filterRows();
				scope.calculatePages(group);

				//this is all for visual purposes to ensure things dont' flash on the screen for .5s while data is loading or being sorted
				if (!scope.rows[group] || scope.rows[group].length <= 0) {
					noDataTimeout = $timeout(function () {
						scope.noData = true;
					}, 300);
				} else {
					scope.noData = false;
					$timeout.cancel(noDataTimeout);
				}

				if (scope.rows[group] && scope.rows[group].length > 0) scope.calculateHeight();

				$timeout(function () {
					scope.registerColumnHover();
				}, 300);
			};

			var noDataTimeout;
			scope.$watch('data', function (newval, oldval) {
				//Logger.debug('ctable', 'Saw table data change ..', newval, oldval);

				if (newval) {
					//Logger.debug('ctable', 'Finished checking multidimensionality', multi);
					if (!scope.multi) {
						//Logger.debug('ctable', 'passing', scope.data, 'to handle data');
						scope.handleData(scope.data, 0);
					} else {
						for (var i in scope.data) {
							//Logger.debug('ctable', 'passing', scope.data[i], 'to handle data');
							scope.handleData(scope.data[i], i);
						}
					}
				} else {
					scope.rows = {};
					scope.mutableRows = {};
				}
			}, true);

			scope.redirect = function (obj) {
				if (!obj['sref']) {
					return;
				}

				Logger.debug('ctable', 'clicking on row happened', obj);

				var link = obj['sref'];
				var key = link['key'];
				var param = {};
				param[key] = link['value'];

				$state.go(link['state'], param);
			};

			scope.$watch('groups', function (newval, oldval) {
				Logger.debug('ctable', 'Saw table groups change ..', newval, oldval);
			}, true);

			//we want to always sort by the "val" attribute of the object, but the user creating the table won't know to include .val
			scope.$watch('orderBy', function (newval, oldval) {
				Logger.debug('ctable', 'Saw orderby change', newval, oldval, scope.orderBy);

				if (scope.orderBy && scope.orderBy.indexOf('.val') < 0) scope.orderBy = scope.orderBy + '.val';

				Logger.debug('ctable', 'Saw orderby change completed', scope.orderBy);
			});

			scope.$watch('settings.page', function (newval, oldval) {
				scope.settings.index = (scope.settings.page - 1) * scope.settings.pageSize;
				Logger.debug('ctable', 'Saw table page change', newval, oldval, scope.settings.index);
			}, true);

			scope.$watch('filters', function (newval, oldval) {
				Logger.debug('ctable', 'Saw table filters change ..', newval, oldval);

				if (typeof oldval !== 'undefined') scope.settings.page = 1; //when filters change, reset the page
				//scope.filterRows();
				scope.calculatePages();
			});

			var loadingTimeout; //make sure this doesn't show up for at least 100ms, so we don't get a lot of jumpyness on loads
			scope.$watch('loading', function (newVal, oldVal) {
				Logger.debug('ctable', 'Saw table loading change ..', newVal, oldVal);
				if (!newVal) {
					$timeout.cancel(loadingTimeout);
					scope.isLoading = false;
					if (scope.data) scope.showContext = scope.data.length <= 0;else scope.showContext = true;
				} else {
					loadingTimeout = $timeout(function () {
						scope.isLoading = true;
					}, 100);
				}
			});
		}
	};
}]);
angular.module('cevo.util').directive('tooltip', ["$sce", function ($sce) {
	return {
		restrict: 'A',
		template: '\n\t\t\t<span class="tooltip" ng-class="[activeClass]">\n\t\t\t\t<ng-transclude></ng-transclude>\n\t\t\t\t<div class="tip-wrapper template" ng-class="[richClass, vPosClass, hPosClass, cPosClass, showClass]">\n\t\t\t\t\t<div class="tip-content" ng-class="[wrapClass]" ng-style="{ \'width\' : (width ? width : \'auto\'), \'text-align\' : (align ? align : \'left\') }">\n\t\t\t\t\t\t<div class="tip-title" ng-if="title">\n\t\t\t\t\t\t\t<i ng-if="!icon" class="ico ico-exclamation-circle"></i>\n\t\t\t\t\t\t\t<i ng-if="icon" class="ico" ng-class="[iconClass]"></i>\n\t\t\t\t\t\t\t<p ng-bind-html="safeTitle" class="nowrap"></p>\n\t\t\t\t\t\t</div>\n\t\t\t\t\t\t<div class="tip" ng-if="safeTooltip">\n\t\t\t\t\t\t\t<div ng-bind-html="safeTooltip"></div>\n\t\t\t\t\t\t</div>\n\t\t\t\t\t\t<div class="tip-footer" ng-if="button && buttonAction">\n\t\t\t\t\t\t\t<a href="" class="button width100p" ng-class="[buttonClass]" ng-bind-html="button" ng-click="buttonAction()"></a>\n\t\t\t\t\t\t</div>\n\t\t\t\t\t</div>\n\t\t\t\t</div>\n\t\t\t</span>',
		scope: {
			title: '@tTitle', //we have to use ttile here just because "title" is an actual html attribute
			tooltip: '@',
			icon: '@tIcon',
			width: '@twidth',
			align: '@talign',
			nowrap: '@tNowrap',
			vposition: '@tVposition',
			hposition: '@tHposition',
			button: '@tButton',
			buttonAction: '&tButtonAction',
			buttonClass: '@tButtonClass',
			active: '@tActive',
			show: '@tShow'
		},
		replace: false, // Replace with the template below
		transclude: true, //'element', // we want to insert custom content inside the directive
		link: function link(scope, elem, attrs) {

			scope.$watch('tooltip', function () {
				scope.safeTooltip = $sce.trustAsHtml(scope.tooltip);
			});

			scope.$watch('title', function () {
				scope.safeTitle = $sce.trustAsHtml(scope.title);
			});

			if (scope.title) scope.richClass = 'rich';

			if (scope.icon) scope.iconClass = 'ico-' + scope.icon;

			if (scope.vposition) scope.vPosClass = 'position-' + scope.vposition;else if (scope.title) scope.vPosClass = 'position-bottom';else scope.vPosClass = 'position-top';

			if (scope.hposition) scope.hPosClass = 'position-' + scope.hposition;else scope.hPosClass = 'position-center';

			if (scope.show) scope.showClass = 'force-show';

			if (scope.nowrap || !scope.width && !scope.title) scope.wrapClass = 'no-wrap';

			if (!scope.buttonClass) scope.buttonClass = 'square outline';

			scope.$watch('active', function (newVal, oldVal) {
				if (typeof scope.active === 'undefined' || scope.active === 'true') scope.active = true;

				if (scope.active === 'false' || scope.active == '') scope.active = false;

				if (scope.active) scope.activeClass = 'active';else scope.activeClass = 'inactive';
			});

			scope.$watch('show', function (newVal, oldVal) {
				if (typeof scope.show === 'undefined' || scope.show === 'false') scope.show = false;
			});
		}
	};
}]);
angular.module('cevo.util').directive('cUpload', ["Logger", function (Logger) {
	return {
		restrict: 'A',
		scope: {
			'cUpload': '=',
			'cUploadProgress': '=?',
			'cUploadType': '@',
			'cUploadHidden': '@'
		},
		link: function link(scope, elem, attrs) {
			scope.cUploadProgress = 0;

			if (typeof scope.cUploadHidden !== 'undefined') {
				elem.addClass('c-upload-hidden');

				elem.parent().on('click', function () {
					this.getElementsByTagName('input')[0].click();
				});
			}

			elem.on('change', function () {
				scope.$apply(function () {
					scope.cUpload = elem[0].files[0];

					if (!scope.cUpload || !scope.cUpload.lastModified) return scope.cUpload = {};

					scope.reader = new FileReader();
					scope.reader.onload = function () {
						scope.$apply(function () {
							scope.cUpload.lastModified = scope.cUpload.lastModified;
							scope.cUpload.lastModifiedDate = scope.cUpload.lastModifiedDate;
							scope.cUpload.name = scope.cUpload.name;
							scope.cUpload.size = scope.cUpload.size;
							scope.cUpload.type = scope.cUpload.type;
							scope.cUpload.src = scope.reader.result;
						});
					};

					scope.reader.onprogress = function (data) {
						scope.data = data;
						scope.cUploadProgress = data.loaded / data.total;
						Logger.debug('cupload', 'onprogress', scope.cUploadProgress);
					};

					if (scope.cUploadType == 'text') {
						scope.reader.readAsText(scope.cUpload);
					} else if (scope.cUploadType == 'buffer') {
						scope.reader.readAsArrayBuffer(scope.cUpload);
					} else if (scope.cUploadType == 'binary') {
						scope.reader.readAsBinaryString(scope.cUpload);
					} else {
						scope.reader.readAsDataURL(scope.cUpload);
					}
				});
			});
		}
	};
}]);
angular.module('cevo.util').directive('cValidate', ["Logger", function (Logger) {
	return {
		restrict: 'A',
		require: 'ngModel',
		link: function link(scope, elem, attr, ngModel) {
			var errors = [],
			    validator;

			var validator = function validator() {
				ngModel.$errors = [];
				ngModel.$setValidity('error', true);
				ngModel.$setPristine();
				return true;
			};

			// add validator to validation loop, that will remove current errors
			ngModel.$validators.validate = validator;

			Logger.debug('form validation', 'linking validate', ngModel);

			scope.registerError = function (error) {
				Logger.debug('form validation', 'registering error', error);
				if (!error) error = { code: 500 };

				errors.push(error.code);
				Logger.debug('form validation', 'pushed new error', error);
				if (!ngModel.$errors) ngModel.$errors = [];

				ngModel.$errors.push(error);
				Logger.debug('form validation', 'set ngmodel $errors', error, ngModel.$setValidity);

				ngModel.$setValidity('error', false);

				Logger.debug('form validation', 'set validity', error);
				return true;
			};

			// expose interface to register API error for this field
			ngModel.$registerError = scope.registerError;
		}
	};
}]);